import os
import shutil
import torch
from ..configs import det_config, segm_config
from ..datasets import get_dataset
from ..models import get_model
from ..losses import get_loss
from ..utils.meters import AverageMeter
from tqdm import tqdm

class Engine:

    def __init__(self, config, device=None):

        self.config = det_config.get_det_config(config) if config.task == 'detection' else segm_config.get_segm_config(config)
        self.device = device if device else torch.device("cuda:0" if torch.cuda.is_available() else "cpu")

        self.setup()

    def setup(self):
        cfg = self.config

        # create output dir
        if cfg['save']:
            os.makedirs(cfg['save_dir'], exist_ok=True)

        self.train_dataset_it, self.val_dataset_it = self.get_dataloaders()
        self.model = self.get_model()
        self.loss_fn = self.get_loss()
        self.optimizer, self.scheduler = self.get_optimizer_and_scheduler()

    def get_dataloaders(self):
        cfg = self.config

        train_dataset = get_dataset(
            cfg['train_dataset']['name'], cfg['train_dataset']['kwargs'])
        train_dataset_it = torch.utils.data.DataLoader(
            train_dataset, batch_size=cfg['train_dataset']['batch_size'], shuffle=True, drop_last=True, num_workers=cfg['train_dataset']['workers'], pin_memory=True if self.device.type == 'cuda' else False)

        # val dataloader
        val_dataset = get_dataset(
            cfg['val_dataset']['name'], cfg['val_dataset']['kwargs'])
        val_dataset_it = torch.utils.data.DataLoader(
            val_dataset, batch_size=cfg['val_dataset']['batch_size'], shuffle=False, drop_last=False, num_workers=cfg['val_dataset']['workers'], pin_memory=True if self.device.type == 'cuda' else False)

        return train_dataset_it, val_dataset_it

    def get_model(self):
        cfg = self.config

        model = get_model(cfg['model']['name'], cfg['model']['kwargs'], cfg['model'].get('init_output')).to(self.device)

        # load checkpoint
        if cfg['weights'] is not None and os.path.exists(cfg['weights']):
            print(f'Loading model from {cfg["weights"]}')
            state = torch.load(cfg['weights'])
            model.load_state_dict(state['model_state_dict'], strict=True)

        return model

    def get_loss(self):
        cfg = self.config

        loss_fn = get_loss(cfg['loss_fn']['name'], cfg['loss_fn'].get('kwargs', {}))

        return loss_fn

    def get_optimizer_and_scheduler(self):
        cfg = self.config

        optimizer = torch.optim.Adam(self.model.parameters(), lr=cfg['lr'])

        lr_lambda = lambda epoch: (1 - (epoch/cfg['n_epochs'])) ** 0.9
        scheduler = torch.optim.lr_scheduler.LambdaLR(optimizer, lr_lambda=lr_lambda)

        return optimizer, scheduler

    def get_metrics(self):
        return {}

    def reset_metrics(self):
        pass

    def compute_metrics(self, pred, sample):
        pass

    def display(self, pred, sample):
        pass

    def forward(self, sample):
        pass

    def trace_model(self):
        pass

    def train_step(self):
        cfg = self.config

        # define meters
        loss_meter = AverageMeter()
 
        self.model.train()
        for i, sample in enumerate(tqdm(self.train_dataset_it)):
            pred, loss = self.forward(sample)

            self.optimizer.zero_grad(set_to_none=True)
            loss.backward()
            self.optimizer.step()

            loss_meter.update(loss.item())

            if cfg['display'] and i % cfg['display_it'] == 0:
                with torch.no_grad():
                    self.display(pred, sample)

        return loss_meter.avg

    def val_step(self):
        cfg = self.config

        self.reset_metrics()
        
        # define meters
        loss_meter = AverageMeter()

        self.model.eval()
        with torch.no_grad():
            for i, sample in enumerate(tqdm(self.val_dataset_it)):
                pred, loss = self.forward(sample)
                loss_meter.update(loss.item())

                if cfg['display'] and i % cfg['display_it'] == 0:
                    self.display(pred, sample)

                self.compute_metrics(pred, sample)

        metrics = self.get_metrics()

        return loss_meter.avg, metrics

    def save_checkpoint(self, epoch, best_val=False, best_val_loss=0, metrics={}):
        cfg = self.config

        state = {
            "epoch": epoch,
            "best_val_loss": best_val_loss,
            "model_state_dict": self.model.state_dict(),
            "optim_state_dict": self.optimizer.state_dict(),
            "metrics": metrics
        }

        print("=> saving checkpoint")
        file_name = os.path.join(cfg['save_dir'], "checkpoint.pth")
        torch.save(state, file_name)
       
        if best_val:
            print("=> saving best_val checkpoint")
            shutil.copyfile(
                file_name, os.path.join(cfg['save_dir'], "best_val_model.pth")
            )

    def train(self):
        cfg = self.config
    
        best_val_loss = float('inf')
        for epoch in range(cfg['n_epochs']):
            print(f'Starting epoch {epoch}')

            train_loss = self.train_step()
            val_loss, metrics = self.val_step()

            print(f'==> train loss: {train_loss}')
            print(f'==> val loss: {val_loss}')
            print(f'metrics: {metrics}')

            best_val = val_loss < best_val_loss
            best_val_loss = min(val_loss, best_val_loss)

            self.save_checkpoint(epoch, best_val=best_val, best_val_loss=best_val_loss, metrics=metrics)

            self.scheduler.step()

        self.trace_model()




import numpy as np

def gaussian2D(shape, sigma=1):
    m, n = [(ss - 1.) / 2. for ss in shape]
    y, x = np.ogrid[-m:m+1,-n:n+1]

    h = np.exp(-(x * x + y * y) / (2 * sigma * sigma))
    h[h < np.finfo(h.dtype).eps * h.max()] = 0
    return h

def get_output_maps(boxes, w, h):

    heatmap = np.zeros((1, h, w), dtype=np.float32)
    scalemap = np.zeros((2, h, w), dtype=np.float32)
    classmap = -1 * np.ones((1, h, w), dtype=np.int64)

    for box in boxes:
        cx, cy = int((box[0] +  box[2]) // 2), int((box[1] + box[3]) // 2)
        bw, bh = int(abs(box[2] - box[0])), int(abs(box[3] - box[1]))
        radius = (min(bw, bh))//2
        diameter = 2 * radius + 1
        gaussian = gaussian2D((diameter, diameter), diameter/6)

        left, right = min(cx, radius), min(w - cx, radius + 1)
        top, bottom = min(cy, radius), min(h - cy, radius + 1)

        masked_heatmap  = heatmap[:, cy - top:cy + bottom, cx - left:cx + right]
        masked_gaussian = gaussian[radius - top:radius + bottom, radius - left:radius + right]
        
        if min(masked_gaussian.shape) > 0 and min(masked_heatmap.shape) > 0:
            heatmap[:, cy - top:cy + bottom, cx - left:cx + right] = np.maximum(masked_heatmap, masked_gaussian)

        # add scale and class
        if cx > 0 and cx < w and cy > 0 and cy < h:
            scalemap[0,cy,cx] = bw
            scalemap[1,cy,cx] = bh
            classmap[0,cy,cx] = box[4]

    return heatmap, scalemap, classmap

def mapping(sample):

    # combinate boxes and category ids
    boxes =  np.array(sample['boxes'])
    category_ids = np.array(sample['category_ids'])
    if len(boxes) > 0:
        boxes_combined = np.concatenate([boxes, category_ids[:,None]], axis=1)
    else:
        boxes_combined = np.array([])

    im_h, im_w = sample['image'].shape[1:]
    heatmap, scalemap, classmap = get_output_maps(boxes_combined, im_w, im_h)

    return {
        'image': sample['image'],
        'boxes': boxes_combined,
        'heatmap': heatmap,
        'scalemap': scalemap,
        'classmap': classmap
    }


import os
import random

from torch.utils.data import Dataset as TorchDataset
import json
import numpy as np

from ml4vision.ml.utils.image_utils import load_image
from PIL import Image


class ML4visionDataset(TorchDataset):
    def __init__(
        self,
        client=None,
        name='',
        owner=None,
        labeled_only=True,
        approved_only=False,
        split=False,
        train=True,
        cache_location="./dataset",
        min_size=None,
        transform=None,
        mapping=None
    ):

        dataset = client.get_dataset_by_name(name, owner=owner)
        dataset.load_samples(labeled_only=labeled_only, approved_only=approved_only)

        if split:
            n_samples = len(dataset.samples)
            dataset.samples = (
                dataset.samples[0 : int(n_samples * 0.8)]
                if train
                else dataset.samples[int(n_samples * 0.8) :]
            )

        dataset_loc = dataset.pull(location=cache_location)

        self.dataset = dataset
        self.dataset_loc = dataset_loc
        self.size = len(self.dataset.samples)
        self.min_size = min_size
        self.transform = transform
        self.mapping = mapping

    def __len__(self):
        return self.min_size or self.size

    def get_index(self, index):
        if self.min_size:
            index = random.randint(0, self.size - 1) if self.size > 1 else 0
        return index

    def get_image(self, sample):
        image_filename = sample.asset["filename"]
        image_path = os.path.join(self.dataset_loc, "images", image_filename)
        image = load_image(image_path)
        return image


class ObjectDetectionDataset(ML4visionDataset):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        print(f"ObjectDetectionDataset created, found {self.size} samples")

    def format_boxes(self, annotations, im_w, im_h):
        boxes = []
        category_ids = []
        for ann in annotations:

            x1, y1, w, h = ann["bbox"]
            x2, y2 = x1 + w - 1, y1 + h - 1

            if x1 < 0:
                x1 = 0
            if y1 < 0:
                y1 = 0

            if x2 >= im_w:
                x2 = im_w - 1
            if y2 >= im_h:
                y2 = im_h - 1

            # remove corrupted boxes
            if x1 >= x2 or y1 >= y2 or w == 0 or h == 0:
                continue

            box = [x1, y1, x2, y2]
            boxes.append(box)
            category_ids.append(ann["category_id"] - 1)

        return boxes, category_ids

    def get_boxes(self, sample, im_w, im_h):
        label_filename = os.path.splitext(sample.asset["filename"])[0] + ".json"
        label_path = os.path.join(self.dataset_loc, "labels", label_filename)
        with open(label_path, "r") as f:
            annotations = json.load(f)["annotations"]
        boxes, category_ids = self.format_boxes(annotations, im_w, im_h)
        return boxes, category_ids

    def __getitem__(self, index):

        index = self.get_index(index)
        sample = self.dataset.samples[index]

        image = np.array(self.get_image(sample))
        im_h, im_w = image.shape[:-1]

        boxes, category_ids = self.get_boxes(sample, im_w, im_h)

        if self.transform:
            transformed = self.transform(
                image=image, bboxes=boxes, category_ids=category_ids
            )
            image = transformed["image"]
            boxes = transformed["bboxes"]
            category_ids = transformed["category_ids"]

        sample = {
            "image": image,
            "boxes": boxes,
            "category_ids": category_ids,
        }

        if self.mapping:
            sample = self.mapping(sample)

        return sample


class SegmentationDataset(ML4visionDataset):
    def __init__(self, *args, ignore_zero=False, **kwargs):
        super().__init__(*args, **kwargs)
        self.ignore_zero = ignore_zero
        print(f"SegmentationDataset created, found {self.size} samples")

    def get_label(self, sample):
        label_filename = os.path.splitext(sample.asset["filename"])[0] + ".png"
        label_path = os.path.join(self.dataset_loc, "labels", label_filename)
        label = np.array(Image.open(label_path))

        if self.ignore_zero:
            label = label - 1

        return label

    def __getitem__(self, index):
        index = self.get_index(index)
        sample = self.dataset.samples[index]

        image = np.array(self.get_image(sample))
        label = self.get_label(sample)

        if self.transform:
            transformed = self.transform(image=image, mask=label)
            image = transformed["image"]
            label = transformed["mask"]

        sample = {
            "image": image,
            "label": label,
        }

        if self.mapping:
            sample = self.mapping(sample)

        return sample


class InstanceSegmentationDataset(ML4visionDataset):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        print(f"InstanceSegmentationDataset created, found {self.size} samples")

    def get_label(self, sample):
        cls_filename = os.path.splitext(sample.asset["filename"])[0] + "_cls.png"
        cls_path = os.path.join(self.dataset_loc, "labels", cls_filename)
        cls = np.array(Image.open(cls_path))

        inst_filename = os.path.splitext(sample.asset["filename"])[0] + "_inst.png"
        inst_path = os.path.join(self.dataset_loc, "labels", inst_filename)
        inst = np.array(Image.open(inst_path))

        return cls, inst

    def __getitem__(self, index):
        index = self.get_index(index)
        sample = self.dataset.samples[index]

        image = np.array(self.get_image(sample))
        cls_label, inst_label = self.get_label(sample)

        if self.transform:
            transformed = self.transform(image=image, masks=[cls_label, inst_label])
            image = transformed["image"]
            cls_label, inst_label = transformed["masks"]

        sample = {"image": image, "cls_label": cls_label, "inst_label": inst_label}

        if self.mapping:
            sample = self.mapping(sample)

        return sample



import brambox as bb
from brambox.io.parser import DetectionParser
from brambox.io.parser import AnnotationParser
import uuid

class APMeter:

    def __init__(self):
        self.detections = DetectionParser()
        self.annotations = AnnotationParser()

    def add_detections(self, pred_boxes, pred_scores, gt_boxes, image_id=None):
        image_id = image_id if image_id else str(uuid.uuid4())
        
        self.detections.append_image(image_id)
        for box, sc in zip(pred_boxes, pred_scores):
            box = box.tolist()
            x, y, w, h = box[0], box[1], box[2]-box[0] + 1, box[3] - box[1] + 1
            self.detections.append(image_id, 
                class_label=box[4],
                x_top_left=x,
                y_top_left=y,
                width=w,
                height=h,
                confidence=sc.item())

        self.annotations.append_image(image_id)
        for box in gt_boxes:
            x, y, w, h = box[0], box[1], box[2]-box[0] + 1, box[3] - box[1] + 1
            self.annotations.append(image_id, 
                class_label=box[4],
                x_top_left=x,
                y_top_left=y,
                width=w,
                height=h)

    def get_ap(self):
        det = self.detections.get_df()
        ann = self.annotations.get_df()
        
        # calculate ap
        pr = bb.stat.pr(det, ann, 0.5)
        ap = bb.stat.ap(pr)

        # calculate best F1
        f1 = bb.stat.fscore(pr)
        threshold = bb.stat.peak(f1)

        # Find point on PR-curve that matches the computed detection threshold
        pr_point = bb.stat.point(pr, threshold.confidence)

        return ap, dict(confidence=threshold.confidence, 
                        f1=threshold.f1, 
                        precision=pr_point.precision, 
                        recall=pr_point.recall)



from .dataset import ObjectDetectionDataset, SegmentationDataset

def get_dataset(name, dataset_kwargs={}):
    if name == "segmentation":
        return SegmentationDataset(**dataset_kwargs)
    elif name == "detection":
        return ObjectDetectionDataset(**dataset_kwargs)
    else:
        raise RuntimeError(f'Dataset {name} is not available!') 


import requests
import os
from urllib.request import urlretrieve
import json
from multiprocessing import Pool
from tqdm import tqdm
from itertools import repeat   
from ml4vision.utils import mask_utils 
from PIL import Image
import numpy as np

class MLModel:

    def __init__(self, client, **kwargs):
        self.client = client
        for key, value in kwargs.items():
            setattr(self, key, value)

    def add_version(self, model_file, params={}):

        # create asset
        filename = os.path.basename(model_file)
        payload = {
            'filename': filename,
        }
        asset_data = self.client.post(f'/assets/', payload=payload)

        # upload file to s3
        url = asset_data['presigned_post_fields']['url']
        fields = asset_data['presigned_post_fields']['fields']
        
        with open(model_file, 'rb') as f:
            response = requests.post(url, data=fields, files={'file':f})
        
        if response.status_code != 204:
            raise Exception(f"Failed uploading to s3, status_code: {response.status_code}")

        # confirm upload
        self.client.put(f'/assets/{asset_data["uuid"]}/confirm_upload/')

        # create version
        payload = {
            'asset': asset_data['uuid'],
            'params': params
        }
        version_data = self.client.post(f'/models/{self.uuid}/versions/', payload)

class Sample:

    def __init__(self, client, **kwargs):
        self.client  = client
        for key, value in kwargs.items():
            setattr(self, key, value)

    def update_label(self, label):
        payload = {
            'annotations': label
        }
        label = self.client.put(f'/samples/{self.uuid}/label/', payload=payload)
        self.label = label

    def load_label(self):
        if self.label is not None:
            sample_details = self.client.get(f'/samples/{self.uuid}/')
            self.label = sample_details['label']
        # return object for multiprocessing
        return self

    def pull_image(self, location='./'):
        asset_filename = self.asset['filename']
        asset_location = os.path.join(location, 'images', asset_filename)
        if not os.path.exists(asset_location):
            urlretrieve(self.asset['url'], asset_location)

    def pull_label(self, location='./', as_json=True, type='BBOX'):
        self.load_label()

        if as_json or type=='BBOX':

            asset_filename = self.asset['filename']
            label_filename = os.path.splitext(asset_filename)[0] + '.json'
            label_location = os.path.join(location, 'labels', label_filename)

            with open(label_location, 'w') as f:
                json.dump(self.label, f)

        elif type == "SEGMENTATION":
            asset_filename = self.asset['filename']
            label_filename = os.path.splitext(asset_filename)[0] + '.png'
            label_location = os.path.join(location, 'labels', label_filename)
            size = self.asset['metadata']['size']

            _, cls = mask_utils.annotations_to_label(self.label['annotations'], size)
            cls.save(label_location)
            
        elif type == "INSTANCE_SEGMENTATION":

            asset_filename = self.asset['filename']
            inst_filename = os.path.splitext(asset_filename)[0] + '_inst.png'
            inst_location = os.path.join(location, 'labels', inst_filename)
            cls_filename = os.path.splitext(asset_filename)[0] + '_cls.png'
            cls_location = os.path.join(location, 'labels', cls_filename)
            size = self.asset['metadata']['size']

            inst, cls = mask_utils.annotations_to_label(self.label['annotations'], size)
            inst.save(inst_location)
            cls.save(cls_location)

        else:
            assert False, f'type {type} is not implemented!'

    def pull(self, location='./', as_json=True, type='BBOX'):
        self.pull_image(location=location)
        self.pull_label(location=location, as_json=as_json, type=type)

    def delete(self):
        self.client.delete(f'/samples/{self.uuid}/')

class Dataset:
    
    def __init__(self, client, **dataset_data):
        self.client = client
        for key, value in dataset_data.items():
            setattr(self, key, value)

    def pull(self, location='./', as_json=False, images_only=False, labels_only=False):
        dataset_loc = os.path.join(location, self.name)

        if format == 'coco':
            image_loc = os.path.join(dataset_loc, 'images')
            os.makedirs(image_loc, exist_ok=True)
            
            print('Loading images')
            with Pool(8) as p:
                inputs = zip(self.samples, repeat(dataset_loc))
                r = p.starmap(Sample.pull_image, tqdm(inputs, total=len(self.samples)))

            print('Loading labels')
            with Pool(8) as p:
                r = p.map(Sample.load_label, tqdm(self.samples, total=len(self.samples)))
                self.samples = r
            
            print('Creating coco json file')
            images = []
            annotations = []

            for sample in self.samples:
                image = {
                    'id': sample.name.rsplit('.', 1)[0],
                    'file_name': sample.name,
                    'width': sample.asset['metadata']['size'][0],
                    'height': sample.asset['metadata']['size'][1]
                }
                images.append(image)

                if sample.label is not None:
                    for i, item in enumerate(sample.label['annotations']):
                        ann = {
                            'id': f"{sample.name.rsplit('.', 1)[0]}_{i}",
                            'image_id': sample.name.rsplit('.', 1)[0],
                            'bbox': item['bbox'],
                            'area': item['area'],
                            'category_id': item['category_id']
                        }
                        annotations.append(ann)

            with open(f'{os.path.join(dataset_loc, self.name)}.json', 'w') as f:
                coco = {
                    'images': images,
                    'categories': self.categories,
                    'annotations': annotations
                }
                json.dump(coco, f)

        else:
            if images_only:
                image_loc = os.path.join(dataset_loc, 'images')
                os.makedirs(image_loc, exist_ok=True)

                print('Downloading your dataset...')
                with Pool(8) as p:
                    inputs = zip(self.samples, repeat(dataset_loc))
                    r = p.starmap(Sample.pull_image, tqdm(inputs, total=len(self.samples)))

                return dataset_loc
            
            elif labels_only:
                label_loc = os.path.join(dataset_loc, 'labels')
                os.makedirs(label_loc, exist_ok=True)

                print('Downloading your dataset...')
                with Pool(8) as p:
                    if as_json:
                        inputs = zip(self.samples, repeat(dataset_loc))
                    else:
                        inputs = zip(self.samples, repeat(dataset_loc), repeat(False), repeat(self.annotation_type))
                    r = p.starmap(Sample.pull_label, tqdm(inputs, total=len(self.samples)))

                return dataset_loc
            
            else:
                image_loc = os.path.join(dataset_loc, 'images')
                label_loc = os.path.join(dataset_loc, 'labels')

                os.makedirs(image_loc, exist_ok=True)
                os.makedirs(label_loc, exist_ok=True)

                print('Downloading your dataset...')
                with Pool(8) as p:
                    if as_json:
                        inputs = zip(self.samples, repeat(dataset_loc))
                    else:
                        inputs = zip(self.samples, repeat(dataset_loc), repeat(False), repeat(self.annotation_type))
                    r = p.starmap(Sample.pull, tqdm(inputs, total=len(self.samples)))

                return dataset_loc

    def push(self, image_list, label_list=None):
        
        print('Uploading data')
        with Pool(8) as p:
            if label_list:
                inputs = zip(repeat(self), image_list, label_list)
            else:
                inputs = zip(repeat(self), image_list)
            r = p.starmap(Dataset.create_sample, tqdm(inputs, total=len(image_list)))

    def load_samples(self, labeled_only=False, approved_only=False):
        samples = []
        
        filter = ''
        if approved_only:
            filter += '&approved=True'
        if labeled_only:
            filter +='&labeled=True'
        
        page=1
        while(True):
            try:
                endpoint = f'/datasets/{self.uuid}/samples/?page={page}'
                endpoint += filter
                for sample in self.client.get(endpoint):
                    samples.append(Sample(self.client, **sample))
                page+=1
            except:
                break
        
        self.samples = samples

    def create_sample(self, image_file, label_file=None, tags={}):
        # create asset
        filename = os.path.basename(image_file)
        payload = {
            'filename': filename,
            'type': 'IMAGE'
        }
        asset_data = self.client.post(f'/assets/', payload=payload)

        # upload file to s3
        url = asset_data['presigned_post_fields']['url']
        fields = asset_data['presigned_post_fields']['fields']
        
        with open(image_file, 'rb') as f:
            response = requests.post(url, data=fields, files={'file':f})
        
        if response.status_code != 204:
            raise Exception(f"Failed uploading to s3, status_code: {response.status_code}")

        # confirm upload
        self.client.put(f'/assets/{asset_data["uuid"]}/confirm_upload/')

        # create sample
        payload = {
            'name': filename,
            'asset': asset_data['uuid'],
            'tags': tags
        }
        sample_data = self.client.post(f'/datasets/{self.uuid}/samples/', payload)

        sample = Sample(client=self.client, **sample_data)

        # upload label
        if label_file:
            label = np.array(Image.open(label_file))
            annotations = mask_utils.label_to_annotations(label)
            sample.update_label(annotations)

        return sample

    def delete(self):
        self.client.delete(f'/datasets/{self.uuid}/')

class Client:

    def __init__(self, apikey, url="https://api.ml4vision.com"):

        self.url = url
        self.apikey = apikey
        self.username, self.email = self.get_owner()

    def get_owner(self):
        owner = self.get('/auth/users/me')
        return owner['username'], owner['email']

    def get(self, endpoint):

        response = requests.get(self.url + endpoint, headers=self._get_headers())
        
        if response.status_code == 200:
            return response.json()
        else:
            raise Exception(f"Request failed, status_code: {response.status_code}")

    def post(self, endpoint, payload={}, files=None):

        response = requests.post(self.url + endpoint, json=payload, files=files, headers=self._get_headers())

        if response.status_code == 201:
            return response.json()
        else:
            raise Exception(f"Request failed, status_code: {response.status_code} - {response.text}")

    def put(self, endpoint, payload={}):

        response = requests.put(self.url + endpoint, json=payload, headers = self._get_headers())

        if response.status_code == 200:
            return response.json()
        else:
            raise Exception(f"Request failed, status_code: {response.status_code} - {response.text}")


    def delete(self, endpoint):
        response = requests.delete(self.url + endpoint, headers=self._get_headers())
        
        if response.status_code != 204:
            raise Exception(f"Request failed, status_code: {response.status_code}")

    def _get_headers(self):

        # set content type & authorization token
        headers = {
            'Authorization': f"APIKey {self.apikey}"
        }

        return headers

    def list_datasets(self):
        datasets = []
        
        page=1
        while(True):
            try:
                for dataset_data in self.get(f'/datasets/?page={page}'):
                    datasets.append(Dataset(self, **dataset_data))
                page+=1
            except:
                break
        
        return datasets

    def get_dataset_by_uuid(self, dataset_uuid):
        dataset_data = self.get(f'/datasets/{dataset_uuid}/')
        return Dataset(self, **dataset_data)

    def get_dataset_by_name(self, name, owner=None):
        owner = owner if owner else self.username
        dataset_data = self.get(f'/datasets/?name={name}&owner={owner}')
        
        if len(dataset_data) == 0:
            raise Exception(f'Did not found dataset "{name}" for owner "{owner}". If this is a shared or public dataset, please specify the owner')

        return Dataset(self, **dataset_data[0])

    def create_dataset(self, name, description='', categories=[{'id': 1, 'name': 'object'}] ,annotation_type='BBOX'):
        payload = {
            'name': name,
            'description': description,
        }
        if categories:
            payload['categories'] = categories
        if annotation_type:
            payload['annotation_type'] = annotation_type

        dataset_data = self.post('/datasets/', payload)
        
        return Dataset(self, **dataset_data)

    def get_or_create_dataset(self, name, owner=None, **kwargs):
        try:
            dataset = self.get_dataset_by_name(name, owner)
        except:
            dataset = self.create_dataset(name, **kwargs)

        return dataset

    def get_model_by_name(self, name, owner=None):
        owner = owner if owner else self.username
        model_data = self.get(f'/models/?name={name}&owner={owner}')
        
        if len(model_data) == 0:
            raise Exception(f'Did not found model "{name}" for owner "{owner}". If this is a shared or public model, please specify the owner')

        return MLModel(self, **model_data[0])


    def create_model(self, name, description='', dataset=None, categories=[] ,annotation_type='BBOX', architecture=''):
        payload = {
            'name': name,
            'description': description,
            'dataset': dataset,
            'categories': categories,
            'annotation_type': annotation_type,
            'architecture': architecture
        }

        model_data = self.post('/models/', payload)

        return MLModel(self, **model_data)

    def get_or_create_model(self, name, owner=None, **kwargs):
        try:
            model = self.get_model_by_name(name,owner=owner)
        except:
            model = self.create_model(name, **kwargs)
        
        return model




from ml4vision.utils.colormap import get_colormap
from ml4vision.utils.rle_cython import compute_rle
from PIL import Image
import numpy as np

def ann_to_mask(ann, size):
    w, h = size
    mask = np.zeros(w * h, dtype=np.uint8)

    index = 0
    zeros = True
    for count in ann['counts']:
        if not zeros:
            mask[index : index + count] = 1
        index+=count
        zeros = not zeros

    return np.reshape(mask, [h, w])


def annotations_to_label(annotations, size):

    w, h = size
    inst = np.zeros((h, w), dtype=np.uint16)
    cls = np.zeros((h, w), dtype=np.uint8)

    for i, ann in enumerate(annotations):
        x, y, w, h = ann['bbox']
        mask = ann_to_mask(ann['segmentation'], (w,h))
        inst[y:y+h,x:x+w] += mask * (i+1) 
        cls[y:y+h,x:x+w] += mask * ann["category_id"]

    cls = Image.fromarray(cls, mode='P')
    cls.putpalette(get_colormap())

    inst = Image.fromarray(inst, mode='I;16')

    return inst, cls

def get_rle(mask):
    counts = compute_rle(mask.flatten())
    size = [mask.shape[0], mask.shape[1]]
    return {"counts": counts, "size": size}

def label_to_annotations(label):

    annotations = []

    # loop over unique instances
    instance_ids = np.unique(label)
    instance_ids = instance_ids[instance_ids != 0]

    for id in instance_ids:
        mask = (label == id)
        
        y, x = np.nonzero(mask)
        if len(y) > 0:
            x_min, x_max, y_min, y_max = int(np.min(x)), int(np.max(x)), int(np.min(y)), int(np.max(y))
            cropped_mask = mask[y_min:y_max+1, x_min:x_max+1]
            rle = get_rle(cropped_mask)

            annotations.append(
                {
                    'segmentation': {
                        'counts': rle['counts'],
                        'size': rle['size']
                    },
                    'bbox': [x_min, y_min, x_max - x_min + 1, y_max - y_min + 1],
                    'category_id': 1
                }
            )

    return annotations

def empty_mask(size):
    return Image.new('L',size)


import albumentations as A
from albumentations.pytorch import ToTensorV2
from ..utils.centernet.mapping import mapping as centernet_mapping

def get_det_config(config):

    engine_config = dict(
        categories = config.dataset.categories,

        save = config.save,
        save_dir = config.save_location,
    
        display = config.display,
        display_it = config.display_it,

        weights = config.weights,

        train_dataset = {
            'name': 'detection',
            'kwargs': {
                'client': config.client,
                'name': config.dataset.name,
                'owner': config.dataset.owner,
                'labeled_only': config.dataset.labeled_only,
                'approved_only': config.dataset.approved_only,
                'split': True,
                'train': True,
                'cache_location': config.dataset.cache_location,
                'min_size': config.dataset.min_size,
                'transform': get_train_transform(config),
                'mapping': centernet_mapping
            },
            'batch_size': config.dataloader.train_batch_size,
            'workers': config.dataloader.train_num_workers
        }, 

        val_dataset = {
            'name': 'detection',
            'kwargs': {
                'client': config.client,
                'name': config.dataset.name,
                'owner': config.dataset.owner,
                'labeled_only': config.dataset.labeled_only,
                'approved_only': config.dataset.approved_only,
                'split': True,
                'train': False,
                'cache_location': config.dataset.cache_location,
                'transform': get_val_transform(config),
                'mapping': centernet_mapping
            },
            'batch_size': config.dataloader.val_batch_size,
            'workers': config.dataloader.val_num_workers
        }, 

        model = {
            'name': config.model.name,
            'kwargs': dict(
                **config.model.kwargs,
                classes = 3 + (len(config.dataset.categories) if len(config.dataset.categories) > 1 else 0)
            ),
            'init_output': True
        },

        loss_fn = {
            'name': 'centernet',
        },

        lr = config.solver.lr,
        n_epochs = config.solver.num_epochs
    )

    return engine_config

def get_train_transform(config):
    transform_list = []
    
    if config.transform.resize:
        transform_list.append(A.SmallestMaxSize(max_size=config.transform.min_size))
    if config.transform.crop:
        crop_size = config.transform.crop_size
        min_size = crop_size - (crop_size * 0.15)
        max_size = crop_size + (crop_size * 0.15)
        transform_list.append(A.RandomSizedCrop([int(min_size),int(max_size)],config.transform.crop_size,config.transform.crop_size))
    if config.transform.flip_horizontal:
        transform_list.append(A.HorizontalFlip(p=0.5))
    if config.transform.flip_vertical:
        transform_list.append(A.VerticalFlip(p=0.5))
    if config.transform.random_brightness_contrast:
        transform_list.append(A.RandomBrightnessContrast(p=0.5))
    
    transform_list.extend([
        A.PadIfNeeded(min_height=None, min_width=None, pad_height_divisor=32,pad_width_divisor=32),
        A.Normalize(),
        ToTensorV2(),
    ])

    transform = A.Compose(
        transform_list, 
        bbox_params=A.BboxParams('pascal_voc', 
        label_fields=['category_ids'], 
        min_visibility=0.1)
    )

    return transform

def get_val_transform(config):
    transform = A.Compose([
        A.SmallestMaxSize(max_size=config.transform.min_size),
        A.PadIfNeeded(min_height=None, min_width=None, pad_height_divisor=32,pad_width_divisor=32),
        A.Normalize(),
        ToTensorV2(),
    ], bbox_params=A.BboxParams('pascal_voc', label_fields=['category_ids'], min_visibility=0.1))

    return transform



import numpy as np

def get_colors(n=256, float=False):
    colors = []

    for i in range(n):
        r, g, b = 0, 0, 0

        for j in range(8):
            r = r + (1 << (7 - j)) * ((i & (1 << (3 * j))) >> (3 * j))
            g = g + (1 << (7 - j)) * ((i & (1 << (3 * j + 1))) >> (3 * j + 1))
            b = b + (1 << (7 - j)) * ((i & (1 << (3 * j + 2))) >> (3 * j + 2))

        colors.append([r, g, b])

    colors = np.array(colors)
    if float:
        colors = colors / 255

    return colors
    



import torch

class IOUEvaluator:

    def __init__(self, n_classes, ignore_index=-1):
        self.n_classes = n_classes
        self.ignore_index = ignore_index
        self.reset()

    def reset (self):
        self.tp = torch.zeros(self.n_classes).double()
        self.fp = torch.zeros(self.n_classes).double()
        self.fn = torch.zeros(self.n_classes).double()        

    # sizes should be "batch_size x 1 x H x W"
    def addBatch(self, x, y):   # x=preds, y=targets
        # clone as we will change x & y if ignore
        x = x.clone()
        y = y.clone()

        if (x.is_cuda or y.is_cuda):
            x = x.cuda()
            y = y.cuda()

        # remove ignore label from target & pred by setting both target and pred to class 0
        ignores = 0
        if (self.ignore_index != -1):
            ignore_mask = (y == self.ignore_index)
            ignores = ignore_mask.sum().item()
            x[ignore_mask] = 0
            y[ignore_mask] = 0

        # scatter pred "batch_size x 1 x H x W"  to onehot
        x_onehot = torch.zeros(
            x.size(0), self.n_classes, x.size(2), x.size(3))
        if x.is_cuda:
            x_onehot = x_onehot.cuda()
        x_onehot.scatter_(1, x, 1).float()

        # scatter target "batch_size x 1 x H x W"  to onehot
        y_onehot = torch.zeros(
            y.size(0), self.n_classes, y.size(2), y.size(3))
        if y.is_cuda:
            y_onehot = y_onehot.cuda()
        y_onehot.scatter_(1, y, 1).float()

        # times prediction and gt coincide is 1
        tpmult = x_onehot * y_onehot
        tp = torch.sum(torch.sum(torch.sum(tpmult, dim=0, keepdim=True), dim=2, keepdim=True), dim=3, keepdim=True).squeeze()
        tp[0] = tp[0] - ignores
        
        # times prediction says its that class and gt says its not
        fpmult = x_onehot * (1-y_onehot) 
        fp = torch.sum(torch.sum(torch.sum(fpmult, dim=0, keepdim=True), dim=2, keepdim=True), dim=3, keepdim=True).squeeze()
        
        # times prediction says its not that class and gt says it is
        fnmult = (1-x_onehot) * (y_onehot)
        fn = torch.sum(torch.sum(torch.sum(fnmult, dim=0, keepdim=True), dim=2, keepdim=True), dim=3, keepdim=True).squeeze()

        self.tp += tp.double().cpu()
        self.fp += fp.double().cpu()
        self.fn += fn.double().cpu()

    def getIoU(self):
        num = self.tp
        den = self.tp + self.fp + self.fn + 1e-15
        iou = num / den
        return torch.mean(iou).item(), list(iou.numpy())     # returns "iou mean", "iou per class"

    def getRecall(self):
        num = self.tp
        den = self.tp + self.fn + 1e-15
        recall = num / den
        return torch.mean(recall).item(), list(recall.numpy())     # returns "recall mean", "recall per class"


import numpy as np
import torch
from .colors import get_colors

class Colorizer:

    def __init__(self):
        self.colors = get_colors()

    def _colorize(self, image_tensor):
        image_np = image_tensor.numpy()

        color_image = np.zeros((3, image_np.shape[0], image_np.shape[1]), np.uint8)

        unique_instances = np.unique(image_np)

        for id in unique_instances:
            id = id % 256
            mask = image_np == id

            color_image[0][mask] = self.colors[id][0]
            color_image[1][mask] = self.colors[id][1]
            color_image[2][mask] = self.colors[id][2]

        return torch.from_numpy(color_image).float()/256

    def colorize(self, image_tensor):
        image_tensor = image_tensor.squeeze()
        if image_tensor.dim() == 3:
            return torch.cat([self._colorize(image_t).unsqueeze(0) for image_t in image_tensor], 0)
        else:
            return self._colorize(image_tensor)


import torch
from segmentation_models_pytorch import Unet



def get_model(name, model_kwargs={}, init_output=False):
    if name == "unet":
        model = Unet(**model_kwargs)
        if init_output:
            with torch.no_grad():
                model.segmentation_head[0].bias[0].fill_(-2.19)
                model.segmentation_head[0].bias[1].fill_(50)
                model.segmentation_head[0].bias[2].fill_(50)
        return model
    else:
        raise RuntimeError("model \"{}\" not available".format(name))




class Node(dict):

    def __init__(self, init_dict=dict()):
        super().__init__(init_dict)

    def __getattr__(self, name):
        if name in self:
            return self[name]
        else:
            raise AttributeError(name)

    def __setattr__(self, name, value):
        self[name] = value

def get_default_config(client, dataset_name, dataset_owner=None):

    dataset = client.get_dataset_by_name(dataset_name, dataset_owner)
    
    if dataset.annotation_type == 'BBOX':
        task = 'detection'
    elif dataset.annotation_type == 'SEGMENTATION':
        task = 'segmentation'
    else:
        raise RuntimeError(f'Trainer not implemented for dataset of annotation type: {dataset.annotation_type}.')
    
    categories = dataset.categories

    cfg = Node(
        dict(
            client = client,

            task = task,

            save = True,
            save_location = './output',

            display = True,
            display_it = 50,

            weights = None,

            dataset = Node(dict(
                name = dataset_name,
                owner = dataset_owner,
                categories = categories,
                labeled_only = True,
                approved_only = False,
                cache_location = './dataset',
                min_size = 1000
            )),

            dataloader = Node(dict(
                train_batch_size = 4,
                val_batch_size = 1,
                train_num_workers = 4,
                val_num_workers = 4
            )),

            model = Node(dict(
                name = 'unet',
                kwargs = dict(
                    encoder_name = 'resnet18',
                    encoder_weights = 'imagenet'
                )
            )),

            solver = Node(dict(
                lr = 5e-4,
                num_epochs = 10
            )),

            transform = Node(dict(
                resize = True,
                min_size = 512,
                crop = True,
                crop_size = 256,
                flip_horizontal = True,
                flip_vertical = True,
                random_brightness_contrast = True,
            ))
        )
    )

    return cfg


from torch import nn
from .bcedice import BCEDiceLoss
from .centernet import centernet_loss


def get_loss(name, loss_kwargs={}):
    if name == 'bce':
        return nn.BCEWithLogitsLoss(**loss_kwargs)
    elif name == 'crossentropy':
        return nn.CrossEntropyLoss(**loss_kwargs)
    elif name == 'bcedice':
        return BCEDiceLoss(**loss_kwargs)
    elif name == 'centernet':
        return centernet_loss
    else:
        raise RuntimeError(f'Loss {name} is not available!')


from torch import nn
from segmentation_models_pytorch.losses import DiceLoss


class BCEDiceLoss(nn.Module):

    def __init__(self, ignore_index=None):
        super().__init__()
        self.ignore_index=ignore_index
        self.bce = nn.BCEWithLogitsLoss()
        self.dice = DiceLoss(mode='binary', from_logits=True, ignore_index=ignore_index)

    def forward(self, y_pr, y_gt):
        if self.ignore_index:
            valid = y_gt != self.ignore_index
            bce_loss = self.bce(y_pr[valid], y_gt[valid])
        else:
            bce_loss = self.bce(y_pr, y_gt)
        return self.dice(y_pr, y_gt) + bce_loss


from .detection_engine import DetectionEngine
from .segmentation_engine import SegmentationEngine

def get_engine(config):
    if config['task'] == 'detection':
        return DetectionEngine(config)
    elif config['task'] == 'segmentation':
        return SegmentationEngine(config)
    else:
        RuntimeError(f'Engine of type {config["task"]} is not available.')


from .engine import Engine
from ..utils.collate import box_collate_fn
from ..utils.visualizer import ObjectDetectionVisualizer
from ..utils.centernet.parse_detections import parse_detections, parse_batch_detections
from ..utils.ap_metrics import APMeter
import torch
import os
import json


class DetectionEngine(Engine):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # set visualizer
        self.visualizer = ObjectDetectionVisualizer()

        # set custom collate function
        self.train_dataset_it.collate_fn = box_collate_fn
        self.val_dataset_it.collate_fn = box_collate_fn

    def reset_metrics(self):
        self.ap_meter = APMeter()

    def compute_metrics(self, pred, sample):
        pred_boxes, pred_scores = parse_batch_detections(pred, threshold=0.1)
        for pred, scores, gt in zip(pred_boxes, pred_scores, sample['boxes']):
            self.ap_meter.add_detections(pred, scores, gt)

    def get_metrics(self):
        ap, ap_dict = self.ap_meter.get_ap()
        return {
            'map': ap,
            'working_point': ap_dict
        }

    def display(self, pred, sample):
        image = sample['image'][0]
        pred_boxes, _ = parse_detections(pred[0], threshold=0.1)
        gt_boxes = sample['boxes'][0]

        self.visualizer.display(image, pred_boxes.cpu(), gt_boxes)

    def forward(self, sample):
        images = sample['image'].to(self.device)
        heatmap = sample['heatmap'].to(self.device)
        scalemap = sample['scalemap'].to(self.device)
        classmap = sample['classmap'].to(self.device)

        pred = self.model(images)
        loss = self.loss_fn(pred, heatmap, scalemap, classmap)
        
        return pred, loss

    def trace_model(self):
        cfg = self.config

        # load best checkpoint
        model = self.model.to(torch.device('cpu'))
        state = torch.load(os.path.join(cfg['save_dir'], 'best_val_model.pth'), map_location=torch.device('cpu'))
        model.load_state_dict(state['model_state_dict'], strict=True)
        model.eval()

        traced_model = torch.jit.trace(model, torch.randn(1, 3, 512, 512))
        traced_model.save(os.path.join(cfg['save_dir'], 'best_val_model.pt'))

        with open(os.path.join(cfg['save_dir'], 'metrics.json'), 'w') as f:
            json.dump(state['metrics'], f)

        with open(os.path.join(cfg['save_dir'], 'categories.json'), 'w') as f:
            json.dump(cfg['categories'], f)




import torch

def box_collate_fn(batch):
    # extract boxes & category_ids from batch
    boxes_batch = [torch.as_tensor(item.pop('boxes')) for item in batch]
    # process batch using normal collate_fn
    others_batch = torch.utils.data._utils.collate.default_collate(batch)
    # append boxes to output
    others_batch['boxes'] = boxes_batch
    
    return others_batch


import torch
from .engine import Engine
from ..utils.visualizer import SegmentationVisualizer
from ..utils.iou_evaluator import IOUEvaluator
import os
import json


class SegmentationEngine(Engine):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.binary = self.config['model']['kwargs']['classes'] == 1
        self.visualizer = SegmentationVisualizer()

        n_classes = self.config['model']['kwargs']['classes']
        ignore_index = self.config['loss_fn'].get('kwargs',{}).get('ignore_index', -1)
        self.iou_meter = IOUEvaluator(n_classes + 1 if self.binary else n_classes,
                                ignore_index=ignore_index)

    def reset_metrics(self):
        self.iou_meter.reset()

    def compute_metrics(self, pred, sample):
        labels = sample['label'].to(self.device)
        labels = labels.unsqueeze(1).long()

        if self.binary:
            self.iou_meter.addBatch((pred > 0).long(), labels)
        else:
            self.iou_meter.addBatch(pred.argmax(dim=1,keepdim=True), labels)

    def get_metrics(self):
        if self.binary:
            iou = self.iou_meter.getIoU()[1][1] 
            return {'mean_iou': iou}
        else:
            miou, iou = self.iou_meter.getIoU()
            return {'mean_iou': miou, 'class_iou': iou}

    def display(self, pred, sample):
        cfg = self.config

        # display a single image
        image = sample['image'][0]
        pred = (pred[0] > 0).cpu() if self.binary else torch.argmax(pred[0],dim=0).cpu()
        gt = sample['label'][0]

        self.visualizer.display(image, pred, gt)

    def forward(self, sample):
        images = sample['image'].to(self.device)
        labels = sample['label'].to(self.device)

        if self.binary:
            labels = labels.unsqueeze(1).float()
        else:
            labels = labels.long()

        pred = self.model(images)
        loss = self.loss_fn(pred, labels)

        return pred, loss

    def trace_model(self):
        cfg = self.config

        # load best checkpoint
        model = self.model.to(torch.device('cpu'))
        state = torch.load(os.path.join(cfg['save_dir'], 'best_val_model.pth'), map_location=torch.device('cpu'))
        model.load_state_dict(state['model_state_dict'], strict=True)
        model.eval()

        traced_model = torch.jit.trace(model, torch.randn(1, 3, 512, 512))
        traced_model.save(os.path.join(cfg['save_dir'], 'best_val_model.pt'))

        with open(os.path.join(cfg['save_dir'], 'metrics.json'), 'w') as f:
            json.dump(state['metrics'], f)

        with open(os.path.join(cfg['save_dir'], 'categories.json'), 'w') as f:
            json.dump(cfg['categories'], f)


from ml4vision.cli_options import Options
from ml4vision import cli_functions
import getpass
from ml4vision import __version__


def main():
    
    args, parser = Options().parse_args()
    exec_command(args, parser)

def exec_command(args, parser):

    if args.command == "authenticate":
        apikey = getpass.getpass(prompt="API key: ", stream=None)
        apikey = apikey.strip()
        if apikey == "":
            print("No API key entered. Consult the ml4vision documentation on how to get your API key.")
            return
        else:
            cli_functions.authenticate(apikey)

    if args.command == "version":
        print(__version__)

    if args.command == "dataset": 
        if args.action == "pull":
            cli_functions.pull_dataset(args.dataset, args.images_only, args.labels_only, args.approved_only, args.labeled_only)
        elif args.action == "push":
            cli_functions.push_to_dataset(args.dataset, args.path, args.label_path)
        elif args.action == "list":
            cli_functions.list_datasets()

        
if __name__ == "__main__":
    main()


import os
from pathlib import Path
from ml4vision.client import Client
import glob
import sys

def _save_apikey(apikey):
    config_path = Path().home() / '.ml4vision' / 'credentials'
    config_path.parent.mkdir(exist_ok=True)

    with open(config_path, 'w') as f:
        f.write(f"apikey={apikey}\n")

def _load_apikey():
    config_path = Path().home() / '.ml4vision' / 'credentials'
    
    try:
        with open(config_path, 'r') as f:
            apikey = f.readline().split('=',1)[1].strip()
            return apikey
    except:
        print('No API Key detected. First authenticate.')
        sys.exit(1)

def _get_client():
    apikey = _load_apikey()
    client = Client(apikey)
    return client

def authenticate(apikey):
    try:
        client = Client(apikey)
        _save_apikey(apikey)
    except:
        print('API Key is invalid.')
        sys.exit(1)

def pull_dataset(name, images_only, labels_only, approved_only, labeled_only):
    client = _get_client() 
    try:   
        dataset = client.get_dataset_by_name(name)
        dataset.load_samples(labeled_only=labeled_only, approved_only=approved_only)
        dataset.pull(images_only=images_only, labels_only=labels_only)
    except Exception as e:
        print(e)
        sys.exit(1)

def push_to_dataset(name, path, label_path=None):
    client = _get_client()
    try:
        dataset = client.get_dataset_by_name(name)
        
        image_files = []
        image_types = (
            '*.png', '*.PNG', 
            '*.jpg','*.JPG', 
            '*.jpeg', '*.JPEG', 
            '*.tiff', '*.TIFF', 
            '*.tif', '*.TIF'
        )
        for t in image_types:
            image_files.extend(glob.glob(os.path.join(path, t)))
        image_files.sort()

        if label_path:
            label_files = []
            label_types = (
                '*.png', '*.PNG',
                '*tif', '*.TIF'
            )
            for t in label_types:
                label_files.extend(glob.glob(os.path.join(label_path, t)))
            label_files.sort()

            assert len(image_files) == len(label_files), f'Number of images ({len(image_files)}) does not equal number of labels ({len(label_files)})'
            dataset.push(image_files, label_files)
        else:
            dataset.push(image_files)


    except Exception as e:
        print(e)
        sys.exit(1)

def list_datasets():
    client = _get_client()
    datasets = client.list_datasets()
    for dts in datasets:
        print(dts.name)


import torch
import torchvision
from torch.nn.functional import max_pool2d


def nms(boxes, scores, classes, threshold=0.1):
    nms_keep = torchvision.ops.nms(boxes, scores, threshold)
    boxes_keep =  torch.index_select(boxes, 0, nms_keep)
    classes_keep = torch.index_select(classes, 0, nms_keep)
    scores_keep = torch.index_select(scores, 0, nms_keep)
    return torch.cat([boxes_keep, classes_keep.unsqueeze(1)], axis=1), scores_keep

def parse_detections(pred, threshold=0.5):

    # get heatmap
    heatmap = pred[0:1, :, :]
    heatmap = torch.sigmoid(heatmap)

    # get scalemap
    scalemap = pred[1:3, :, :]

    heatmap_max = max_pool2d(heatmap, (7, 7), stride=1, padding=3)  # 1 x h x w
    keep = (heatmap_max == heatmap) * (heatmap > threshold)  # 1 x h x w

    heatmap = heatmap * keep  # 1 x h x w
    cycx = torch.nonzero(heatmap.squeeze())  # 2 x N
    scales = scalemap[keep.expand_as(scalemap)].view(2, -1)  # 2 x N

    w = scales[0, :]  # N
    h = scales[1, :]  # N

    x0 = cycx[:, 1] - w / 2  # N
    y0 = cycx[:, 0] - h / 2  # N

    x1 = cycx[:, 1] + w / 2  # N
    y1 = cycx[:, 0] + h / 2  # N

    boxes = torch.stack((x0, y0, x1, y1), dim=1)
    scores = heatmap[keep]
    if pred.size(0) > 3:
        classmap = torch.argmax(pred[3:,:,:],dim=0, keepdim=True)
        classes = classmap[keep]
    else:
        classes = torch.zeros_like(scores)

    return nms(boxes, scores, classes, threshold=0.5)


def parse_batch_detections(preds, threshold=0.5):
    detections, scores = [], []
    for pred in preds:
        det, sc = parse_detections(pred, threshold=threshold)
        detections.append(det)
        scores.append(sc)
    return detections, scores


import os
import json

def upload_model(config, model_name=None):

    model_name = model_name or f"{config.dataset.name}-model"
    annotation_type = "BBOX" if config.task == 'detection' else "SEGMENTATION"
    architecture = 'object_detection_fn' if annotation_type == "BBOX" else "segmentation_fn"

    model = config.client.get_or_create_model(
        model_name,
        description='',
        dataset = config.client.get_dataset_by_name(config.dataset.name, config.dataset.owner).uuid,
        categories=config.dataset.categories,
        annotation_type=annotation_type,
        architecture=architecture
    )

    with open(os.path.join(config.save_location, 'metrics.json'), 'r') as f:
        metrics = json.load(f)

    print('adding version')
    if annotation_type == "BBOX":
            
        model.add_version(
            os.path.join(config.save_location, 'best_val_model.pt'),
            params = {
                'min_size': config.transform.min_size,
                'pad': 32,
                'normalize': True,
                'threshold': float(metrics["working_point"]["confidence"])
            }
        )

    else:
        model.add_version(
            os.path.join(config.save_location, 'best_val_model.pt'),
            params = {
                'min_size': config.transform.min_size,
                'pad': 32,
                'normalize': True,
            }
        )




import torch
from torch.nn import functional as F

def centernet_loss(pred, heatmap, scalemap, classmap):
    heatmap_loss = sigmoid_focal_loss(pred[:,0:1,:,:], heatmap==1, heatmap)
    scale_loss = l1_loss(pred[:,1:3,:,:], scalemap)
    if pred.size(1) > 3:
        class_loss = F.cross_entropy(pred[:,3:,:,:], classmap.squeeze(1), ignore_index=-1)
        return heatmap_loss + 0.1*scale_loss + class_loss
    else:
        return heatmap_loss + 0.1*scale_loss

def l1_loss(pred, gt):
    valid = gt > 0
    n_points = valid.sum()//2
    loss = 0
    if n_points > 0:
        loss = abs(pred[valid] - gt[valid]).sum() / n_points
    
    return loss

def sigmoid_focal_loss(inputs, targets, alpha):
    inputs = inputs.float()
    targets = targets.float()
    p = torch.sigmoid(inputs)
    ce_loss = F.binary_cross_entropy_with_logits(inputs, targets, reduction="none")
    p_t = p * targets + (1 - p) * (1 - targets)
    loss = ce_loss * ((1 - p_t) ** 2)
    
    alpha_t = alpha * targets + (1 - alpha) * (1 - targets)
    loss = torch.pow(alpha_t, 4) * loss
    loss = loss.sum()

    n_targets = targets.sum()

    if n_targets > 0:
        loss = loss / n_targets 

    return loss



import matplotlib
import matplotlib.pyplot as plt
from .image_utils import denormalize_img_tensor
from .colorizer import Colorizer
from .colors import get_colors
from matplotlib.patches import Rectangle
from IPython import display as idisplay


class SegmentationVisualizer:

    def __init__(self, im_norm=True):
        plt.ion()
        self.inline = 'inline' in matplotlib.get_backend()
        self.dh = None
        self.im_norm = im_norm
        self.colorizer = Colorizer()
        self.window = plt.subplots(ncols=3)

    def display(self, image, pred, gt):
        fig, axs = self.window
        for ax in axs:
            ax.cla()
            ax.set_axis_off()

        # display image
        if self.im_norm:
            image = image.squeeze_()
            image = denormalize_img_tensor(image)
        
        axs[0].imshow(image.numpy().transpose(1,2,0))

        # display pred
        colored_pred = self.colorizer.colorize(pred)
        axs[1].imshow(colored_pred.numpy().transpose(1,2,0))

        # display gt
        colored_gt = self.colorizer.colorize(gt)
        axs[2].imshow(colored_gt.numpy().transpose(1,2,0))

        if self.inline: 
            if self.dh:
                self.dh.update(fig)
            else:
                self.dh = idisplay.display(fig, display_id=True)
        else:
            plt.draw()
        
        self.mypause(0.001)

    @staticmethod
    def mypause(interval):
        backend = plt.rcParams['backend']
        if backend in matplotlib.rcsetup.interactive_bk:
            figManager = matplotlib._pylab_helpers.Gcf.get_active()
            if figManager is not None:
                canvas = figManager.canvas
                if canvas.figure.stale:
                    canvas.draw()
                canvas.start_event_loop(interval)
                return

class ObjectDetectionVisualizer:

    def __init__(self, im_norm=True):
        plt.ion()
        self.inline = 'inline' in matplotlib.get_backend()
        self.dh = None
        self.im_norm = im_norm
        self.colors = get_colors(float=True)[1:]
        self.window = plt.subplots(ncols=2)

    def display(self, image, pred, gt):
        fig, axs = self.window
        for ax in axs:
            ax.cla()
            ax.set_axis_off()

        if self.im_norm:
            image = image.squeeze_()
            image = denormalize_img_tensor(image)

        # display pred
        axs[0].imshow(image.numpy().transpose(1,2,0))
        for box in pred.numpy():
            rect = Rectangle((box[0], box[1]), box[2]-box[0], box[3]-box[1], linewidth=1, edgecolor=self.colors[int(box[4])], facecolor='none')
            axs[0].add_patch(rect)

        # display gt
        axs[1].imshow(image.numpy().transpose(1,2,0))
        for box in gt.numpy():
            rect = Rectangle((box[0], box[1]), box[2]-box[0], box[3]-box[1], linewidth=1, edgecolor=self.colors[int(box[4])], facecolor='none')
            axs[1].add_patch(rect)

        if self.inline: 
            if self.dh:
                self.dh.update(fig)
            else:
                self.dh = idisplay.display(fig, display_id=True)
        else:
            plt.draw()
            
        self.mypause(0.001)

    @staticmethod
    def mypause(interval):
        backend = plt.rcParams['backend']
        if backend in matplotlib.rcsetup.interactive_bk:
            figManager = matplotlib._pylab_helpers.Gcf.get_active()
            if figManager is not None:
                canvas = figManager.canvas
                if canvas.figure.stale:
                    canvas.draw()
                canvas.start_event_loop(interval)
                return







from PIL import Image, ExifTags
import torch

def handle_exif_rotation(image):
    try:
        for orientation in ExifTags.TAGS.keys():
            if ExifTags.TAGS[orientation]=='Orientation':
                break
        exif=dict(image._getexif().items())

        if exif[orientation] == 3:
            image=image.transpose(Image.ROTATE_180)
        elif exif[orientation] == 6:
            image=image.transpose(Image.ROTATE_270)
        elif exif[orientation] == 8:
            image=image.transpose(Image.ROTATE_90)
        return image
    except (AttributeError, KeyError, IndexError):
        return image

def load_image(path, rgb=True):
    image = Image.open(path)
    image = handle_exif_rotation(image)

    if rgb and image.mode != 'rgb':
        image = image.convert('RGB')

    return image

def denormalize_img_tensor(img, mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225]):
    mean = torch.as_tensor(mean, dtype=img.dtype, device=img.device)
    std = torch.as_tensor(std, dtype=img.dtype, device=img.device)

    return img.mul(std[:,None,None]).add(mean[:,None,None])


import sys
from argparse import ArgumentParser
import argcomplete

class Options:

    def __init__(self):

        self.parser = ArgumentParser(
            description="Cli for ml4vision."
        )

        subparsers = self.parser.add_subparsers(dest="command")

        # authenticate
        subparsers.add_parser("authenticate", help="Store apikey for authentication.")

        # dataset
        dataset = subparsers.add_parser("dataset", help="dataset related functions")
        dataset_action = dataset.add_subparsers(dest="action")

        ## dataset list
        dataset_list = dataset_action.add_parser("list", help="List all datasets")
        
        ## dataset pull
        dataset_pull = dataset_action.add_parser("pull", help="Pull samples from dataset")
        dataset_pull.add_argument(
            "dataset",
            type=str,
            help="Name of dataset"
        )
        dataset_pull.add_argument(
            "--images-only",
            action="store_true",
            help="Pull the images only"
        )
        dataset_pull.add_argument(
            "--labels-only",
            action="store_true",
            help="Pull the labels only"
        )
        dataset_pull.add_argument(
            "--approved-only",
            action="store_true",
            help="Pull only the approved samples"
        )
        dataset_pull.add_argument(
            "--labeled-only",
            action="store_true",
            help="Pull only the labeled samples"
        )

        ## dataset push
        dataset_push = dataset_action.add_parser("push", help="Push images to dataset")
        dataset_push.add_argument(
            "dataset",
            type=str,
            help="Name of dataset"
        )
        dataset_push.add_argument(
            "path",
            type=str,
            help="Path to image folder"
        )
        dataset_push.add_argument(
            "--label_path",
            type=str,
            help="Path to label folder"
        )

        # version
        subparsers.add_parser("version", help="Print current version number")

        argcomplete.autocomplete(self.parser)

    def parse_args(self):
        args = self.parser.parse_args()

        if not args.command:
            self.parser.print_help()
            sys.exit()

        return args, self.parser


import re
from pathlib import Path

from setuptools import Extension, setup
from Cython.Build import cythonize
import numpy as np

with open("README.md", "rb") as f:
    long_description = f.read().decode("utf-8")

with open(Path(__file__).parent / "ml4vision" / "__init__.py", "r") as f:
    content = f.read()
    version = re.search(r'__version__\s*=\s*[\'"]([^\'"]*)[\'"]', content).group(1)

extensions = [
    Extension(name="ml4vision.utils.*",
        sources=["ml4vision/utils/*.pyx"],
        libraries=["m"],
        extra_compile_args=["-ffast-math"],
        include_dirs=[np.get_include()])
]

setup(
    name="ml4vision-py",
    version=version,
    author="ml4vision",
    author_email="info@ml4vision.com",
    description="Python sdk and cli for ml4vision",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/ml4vision/ml4vision-py",
    setup_requires=["wheel","Cython","numpy"],
    install_requires=[
        "Cython",
        "tqdm",
        "requests",
        "argcomplete",
        "numpy",
        "pillow"
    ],
    extras_require={
        "ml": [
            "torch>=1.9.0",
            "torchvision>=0.10.0",
            "segmentation_models_pytorch>=0.2.1",
            "opencv-python-headless<4.3", #required on colab
            "albumentations>=1.1.0",
            "brambox>=4.0.0"
        ],
    },
    packages=[
        "ml4vision",
        "ml4vision.utils",
        "ml4vision.ml",
        "ml4vision.ml.configs",
        "ml4vision.ml.datasets",
        "ml4vision.ml.engines",
        "ml4vision.ml.losses",
        "ml4vision.ml.models",
        "ml4vision.ml.utils",
        "ml4vision.ml.utils.centernet",

    ],
    ext_modules=cythonize(extensions),
    classifiers=["Programming Language :: Python :: 3", "License :: OSI Approved :: MIT License"],
    entry_points={"console_scripts": ["ml4vision=ml4vision.cli:main"]},
    python_requires=">=3.6",
    package_data={"ml4vision.utils":["*.pyx"]}
)


class AverageMeter:
    def __init__(
        self,
    ):
        self.reset()

    def reset(self):
        self.sum = 0.0
        self.count = 0.0
        self.avg = 0.0

    def update(self, val):
        self.sum += val
        self.count += 1
        self.avg = self.sum / (self.count + 1e-3)
